/* --------------------------------------------------------------------------
 *
 *  WP
 *
 * -------------------------------------------------------------------------- */

#include "wpj.h"

extern int KARMA_DEF_INIT;
extern int KARMA_DEF_MAX; 
extern int KARMA_DEF_INC; 
extern int KARMA_DEF_DEC;
extern int KARMA_DEF_PENALTY;
extern int KARMA_DEF_RESTORE;
extern int DEF_RATE_TIME;
extern int DEF_RATE_CONN;
extern int DEF_CONN;

/* main structure pointer */
wpjs wpj;
int wpj_shutdown = 0;

void signal_end(int sig)
{
    wpj_shutdown = 1;
}


/* connect and alloc */ 
void connect_serwers()
{
  server_info si;
  wp_connection conn;
  int number = 1;

  wpj->cfg->si = NULL;
  if ( (conn = wpj->cfg->connections) != NULL )
    {
      /* alloc */
      si = pmalloco(wpj->p,sizeof(_server_info));
      memset(si,0,sizeof(_server_info));
      if (j_strcmp(conn->mhost,wpj->cfg->reg_host)==0)
      {
	  wpj->cfg->si = si;
	  log_debug("Reg server is ","%s",wpj->cfg->reg_host);
      }
      si->servertime[0] = 0;
      si->secret = conn->secret;
      si->mhost  = conn->mhost; 
      si->mport  = conn->mport;
      si->mip    = conn->mip;
      si->state  = conn_CLOSED;
      si->number = number;
      number++;

      /* build si ring */
      //      if (wpj->si == NULL)
	wpj->si = si;
	//      else
	//	prev->next = si;

	//      prev = si;

	/* connect */
      io_connect(si->mip, si->mport, server_process_xml, (void*)si,0);
    }
}

/* Allow configuration file */
int configure(char *cfgfile)
{
    xmlnode xcfg, tmp;
    wp_connection conn,prev; /* tmp connection */


    xcfg = xmlnode_file(cfgfile);
    if( !xcfg)
    {
        log_alert("Error","Unable to parse config file: %s", cfgfile);
        return 0;
    }

    /* Our hostname */
    wpj->cfg->hostname = pstrdup(wpj->p, xmlnode_get_tag_data(xcfg, "hostname"));
    if(!wpj->cfg->hostname)
    {
        log_alert("Error","Required hostname not supplied in config.");
        return 0;
    }

    wpj->cfg->reg_host = pstrdup(wpj->p, xmlnode_get_tag_data(xcfg, "register_hostname"));
    if(!wpj->cfg->reg_host)
    {
        log_alert("Error","Required register hostname not supplied in config.");
        return 0;
    }

    /* Our pidfile */
    tmp = xmlnode_get_tag(xcfg, "pidfile");
    if(tmp != NULL)
    {
      wpj->cfg->pidfile = pstrdup(wpj->p, xmlnode_get_tag_data(xcfg, "pidfile"));
    }
    else
      wpj->cfg->pidfile =  "pidfile";

    tmp = xmlnode_get_tag(xcfg, "logfile");
    if(tmp != NULL)
    {
      wpj->cfg->logfile = pstrdup(wpj->p, xmlnode_get_tag_data(xcfg, "logfile"));
    }
    else
      wpj->cfg->logfile =  "log_";

    tmp = xmlnode_get_tag(xcfg, "statfile");
    if(tmp != NULL)
    {
      wpj->cfg->statfile = pstrdup(wpj->p, xmlnode_get_tag_data(xcfg, "statfile"));
    }
    else
      wpj->cfg->statfile =  "liczba_polaczen";


    /* ping time */
    wpj->cfg->ping_time = j_atoi(xmlnode_get_tag_data(xcfg, "pingtime"), DEF_PINGTIME);

    /* iocheck */
    wpj->cfg->iocheck = j_atoi(xmlnode_get_tag_data(xcfg, "pingcheck"), DEF_IOCHECK);

    /* ioclose */
    wpj->cfg->ioclose = j_atoi(xmlnode_get_tag_data(xcfg, "pingclose"), DEF_IOCLOSE);

    /* ioclose */
    wpj->cfg->auth_time = 
      j_atoi(xmlnode_get_tag_data(xcfg, "authtime"), DEF_AUTHTIME);

    /* ioclose */
    wpj->cfg->datacheck = 
      j_atoi(xmlnode_get_tag_data(xcfg, "datacheck"), DEF_DATACHECK);

    /* ioclose */
    wpj->cfg->reconnect_time = 
      j_atoi(xmlnode_get_tag_data(xcfg, "reconnecttime"), DEF_RECONNECTTIME);

    
    /* Setup our listener info */
    tmp = xmlnode_get_tag(xcfg, "listen");
    wpj->cfg->lport = j_atoi(xmlnode_get_tag_data(tmp, "port"), 5222);
    wpj->cfg->lip = pstrdup(wpj->p, xmlnode_get_tag_data(tmp, "ip"));
    if(!wpj->cfg->lip)
    {
        log_alert("","No hostname or IP was specified to listen on.");
        return 0;
    }


    /* Config connection info */
    tmp = xmlnode_get_tag(xcfg, "connectlist");

    /* Czy po padzie serwera automatycznie odnawiac sesje uzytkownikow ? */
    wpj->cfg->auto_refresh_sessions = xmlnode_get_tag(tmp,"auto_refresh_sessions")?1:0;

    log_debug("","cfg autor refresh %d %p",wpj->cfg->auto_refresh_sessions,
	      xmlnode_get_tag(tmp,"auto_refresh_sessions"));

    tmp = xmlnode_get_tag(tmp,"connect");

    while (tmp != NULL)
    {

      if ( xmlnode_get_type(tmp) != NTYPE_TAG )
      {
	tmp = xmlnode_get_nextsibling(tmp);
	continue;
      }

      conn = pmalloco(wpj->p, sizeof(_wp_connection));
      conn->mhost = pstrdup(wpj->p, xmlnode_get_tag_data(tmp, "host"));
      conn->mip = pstrdup(wpj->p, xmlnode_get_tag_data(tmp, "ip"));
      conn->secret = pstrdup(wpj->p, xmlnode_get_tag_data(tmp, "secret"));
      conn->mport = j_atoi(xmlnode_get_tag_data(tmp, "port"), 5225);

      if(!conn->mhost || !conn->secret  || !conn->mip)
      {
	log_alert("Error","Server hostname or secret or ip were not specified.");
	return 0;
      }

      log_debug("","connect %s %s %d",conn->mhost,conn->mip,conn->mport);

      wpj->cfg->conn_number++;
      if (wpj->cfg->connections == NULL) 
	wpj->cfg->connections = conn;
      else
	prev->next = conn;

      prev = conn;
      tmp = xmlnode_get_nextsibling(tmp);
    }



    /* karma config */
    tmp = xmlnode_get_tag(xcfg, "karma");
    if (tmp != NULL)
      {
	KARMA_DEF_INIT    = j_atoi(xmlnode_get_tag_data(tmp, "init"), KARMA_INIT);
	KARMA_DEF_MAX     = j_atoi(xmlnode_get_tag_data(tmp, "max"), KARMA_MAX);
	KARMA_DEF_INC     = j_atoi(xmlnode_get_tag_data(tmp, "inc"), KARMA_INC);
	KARMA_DEF_DEC     = j_atoi(xmlnode_get_tag_data(tmp, "dec"), KARMA_DEC);
	KARMA_DEF_PENALTY = j_atoi(xmlnode_get_tag_data(tmp, "penalty"), KARMA_PENALTY);
	KARMA_DEF_RESTORE = j_atoi(xmlnode_get_tag_data(tmp, "restore"), KARMA_RESTORE);

	log_debug("","KARMA_DEF_INIT %d",KARMA_DEF_INIT);
	log_debug("","KARMA_DEF_MAX %d",KARMA_DEF_MAX);
	log_debug("","KARMA_DEF_INC %d",KARMA_DEF_INC);
	log_debug("","KARMA_DEF_DEC %d",KARMA_DEF_DEC);
	log_debug("","KARMA_DEF_PENALTY %d",KARMA_DEF_PENALTY);
	log_debug("","KARMA_DEF_RESTORE %d",KARMA_DEF_RESTORE);

	wpj->cfg->karma_time = j_atoi(xmlnode_get_tag_data(tmp, "heartbeat"), 2);
	log_debug("","cfg->karma_time  %d",wpj->cfg->karma_time);

      }

    tmp = xmlnode_get_tag(xcfg, "rate");
    if (tmp != NULL)
      {
	DEF_RATE_TIME  = j_atoi(xmlnode_get_attrib(tmp, "time"), 1);
	DEF_RATE_CONN  = j_atoi(xmlnode_get_attrib(tmp, "points"), 2);
	DEF_CONN       = j_atoi(xmlnode_get_attrib(tmp, "count"), 0);

	log_debug("","DEF_RATE_T %d",DEF_RATE_TIME);
	log_debug("","DEF_RATE_P %d",DEF_RATE_CONN);
	log_debug("","DEF_CONN %d",DEF_CONN);
      }

    wpj->cfg->max_clients = j_atoi(xmlnode_get_tag_data(xcfg, "max_clients"), DEF_MAX_CLIENTS-1);
    if (wpj->cfg->max_clients > DEF_MAX_CLIENTS-1) wpj->cfg->max_clients = DEF_MAX_CLIENTS-1;

    log_debug("MAX_CLIENTS","%d",wpj->cfg->max_clients);

    xmlnode_free(xcfg);
    return 1;
}


int main (int argc, char** argv)
{
    int          j,help;
    char         *c;
    char         *cfgfile=NULL;
    int          ret;
    pool         p;
    client_c     ci;

    /* Command line */
    help = 0;
    for(j = 1; j < argc; j++)
    {
        if(argv[j][0]!='-')
        {   /* make sure it's a valid command */
            help=1;
            break;
        }
	
	c=argv[j]+1;
	if(*c == 'c')
	{
	    j++;
	    cfgfile=argv[j];
	    continue;
	}

	if(*c == 'D')
	{
	    debug_flag = 1;
	    continue;
	}

	if(*c == 'V' || *c == 'v')
	{
	    fprintf(stdout,"%s\n",VERSION);
	    exit(0);
	}
      
    }
    
    if(help)
    {
        fprintf(stdout,"USAGE:  %s -c\t\t configuration file -D\t\tenable debug -v\t\tserver version\n -V\t\tserver version\n\n", argv[0]);
        exit(0);
    }


    /* alloc main structure */
    p = pool_new();
    wpj = pmalloco(p, sizeof(_wpjs));
    wpj->p = p;

    /* process config file */
    wpj->cfg = pmalloco(wpj->p, sizeof(_config));
    ret = configure( cfgfile ? cfgfile : "wpj.xml");    
    if (ret == 0)
    {
        log_debug("Error","Could not configure client, exitting!");
	pool_free(wpj->p);
        exit(1);
    }

    /* init log */
    init_log(wpj->cfg->logfile);

    /* init connections */
    for (j = 0 ; j < DEF_MAX_CLIENTS ; j++)
    {
	ci = &(wpj->clients[j]);

	ci->m = NULL;
	ci->next = NULL;
	ci->loc = j;
	
	if (wpj->last_free_client == NULL)
	  wpj->last_free_client = ci;
	else
	  wpj->first_free_client->next = ci;
	
	wpj->first_free_client = ci;	
    }

    /* create pidfile */
    {
      int fd;
      char pidstr[10];
      fd = open (wpj->cfg->pidfile,O_CREAT | O_EXCL | O_WRONLY ,0600);
      if (fd < 0)
	{
	  if (errno == EEXIST)
	    {
	      fprintf(stderr,"Pidfile exist. Remove it first.");
	      pool_free(wpj->p);
	      exit(1);	    
	    }
	  fprintf(stderr,"Error while openning pidfile.");
	  pool_free(wpj->p);
	  exit(1);	    
	}
      sprintf(pidstr,"%d",getpid());
      write(fd,pidstr,strlen(pidstr));
      close(fd);
    }

    log_alert(wpj->cfg->hostname,"...___---:::=== Starting up ===:::---___...");

    signal(SIGPIPE, SIG_IGN);
    signal(2, signal_end);
    signal(3, signal_end);
    signal(SIGTERM, signal_end);

    /* init IO */
    io_init();

    /* connect */
    connect_serwers();

    /* wait for connections */
    sleep(1);

    /* start listening */
    io_listen(wpj->cfg->lport, wpj->cfg->lip, client_process_xml, NULL);


    log_debug(ZONE,"Listening connections");

    while (wpj_shutdown != 1) {
      Sleep(100);
    }

    log_alert("Signal","Recieved Kill. Shutting down...");

    /* stop */
    io_stop();

    /* remove pid file */
    if(wpj->cfg->pidfile != NULL)
    {
         unlink(wpj->cfg->pidfile);
    }

    pool_free(wpj->p);

    stop_log();

    return 0;
}

DWORD timeGetTime()
{
	struct timeb tb;
	DWORD ret;
	ftime(&tb);
	ret=tb.time*1000+tb.millitm;
	return ret;
}

DWORD timeGetTimeSec100()
{
	struct timeb tb;
	DWORD ret;
	ftime(&tb);
	ret = tb.time*100 + (tb.millitm / 8) ; /* zmiana raz na 1/100 sek */
	return ret;
}

DWORD timeGetTimeSec()
{
	struct timeval tv;	
	DWORD ret;
	gettimeofday(&tv,NULL);
	ret = tv.tv_sec;
	return ret;
}

void Sleep(DWORD time)
{
	usleep(time*1000);
}
